﻿#Requires -version 3.0
Set-StrictMode -Version Latest
# $Revision: 434 $

function Test-FolderAccess([IO.DirectoryInfo] $Path)
{
	try 
	{
		$testPath = Join-Path $Path ([IO.Path]::GetRandomFileName())
        New-Item -Type File $testpath -ErrorAction Stop
		Remove-Item $testPath -force
		return $true
    } 
	catch { return $false }
}

Function Remove-FolderUsingRobocopy
{
<#
.SYNOPSIS
    Accepts an array of folders and deletes them all using robocopy as robocopy can delete files and paths longer than 260 characters

.DESCRIPTION
	Accepts one or more folders, including from the pipeline and will use robocopy to remove all of the folders.
	This command is useful specifically in situations where you want to delete one or more folders that contain a total file or path that is greater than 260 characters

.PARAMETER $InputFolders
    A folder or folders to delete.  Can be comma separated or provided by the pipeline

.PARAMETER LogFile
    Optional.  Provide a path save all output of every deleted file (and some basic statistics) to a file for record keeping

.PARAMETER Force
    Bypass confirmation prompts on delete operations.  Use with caution.

.PARAMETER $SkipRobocopyVersionCheck
    Skip by default will check that the version of robocopy available matches those tested.  It's likely however other versions will still work
	so it should be safe to override this.  This is just to make sure you are aware it is untested.

.EXAMPLE
    Remove-FolderUsingRobocopy c:\badfolder -whatif
    This will write to the console every file that will be deleted if this command is executed without whatif
	
.EXAMPLE
    Remove-FolderUsingRobocopy c:\badfolder, c:\badfolder2
	This will permenantly delete these two folders, but will prompt for confirmation for each

.EXAMPLE
    Remove-FolderUsingRobocopy c:\badfolder, c:\badfolder2 -Force -LogFile c:\logfiles\log1.txt
	This will permenantly delete these two folders and will NOT ask for confirmation first.  
	It will write the files deleted both to the screen and to the log file specified

.EXAMPLE
    (Import-CSV c:\temp\folders.csv).Directories | Remove-FoldersWithRobocopy -Force
	Will remove all folders specified in te CSV that have the column name Directories

.NOTES
    Author: Robert Vance
    Date:   December 15th, 2014
#>
	
	[CmdletBinding(SupportsShouldProcess=$true, ConfirmImpact="High")]
	param
	(
		[parameter(Position=0, Mandatory=$true,ValueFromPipeline=$true,
		HelpMessage = "Folder to be deleted. Accepts pipeline input and comma separated list of folders.")]
		[PSObject[]]$InputFolder,

		[parameter(Position=1, Mandatory=$false,
		HelpMessage = "(Optional) Write all files and folders deleted to a log file for historical record keeping.")] 
		[String]
		$LogFile,
		
		[parameter(Mandatory=$false,
		HelpMessage = "Bypass confirmation prompts on delete operations.  Use with caution")] 
		[Switch]$Force,
		
		[parameter(Mandatory=$false,
		HelpMessage = "Bypass version validation for robocopy.exe. The script will likely still work.  This just means it wasn't tested with this version")] 
		[Switch]$SkipRobocopyVersionCheck
	)

	BEGIN
	{
		#Region Variable Declaration
		$ArrayFolder = @()		        # Since this cmdlet accepts both paramater and pipeline input, I need to count the objects in the pipeline first in the PROCESS block
		$BadFolders = @()				# Any folders requested for removal that did not exist at the time of execution
		$WhatIfPrefix = $null			# Adds "[Whatif]" to all log entries when -whatif is enabled
		$RejectAll = $false				# Used to add the Yes to All and No to All functionality to the Log overwrite/append confirmation dialog
		$ConfirmAll = $false
		$IsFolder = -1					# When Robocopy returns a folder, it has a size of -1.  This is saved as a variable to make it more readable
		$FolderSize = 0					# Size in bytes of each folder requested
		$AllFoldersSize = 0				# Running total Size in bytes of the total size of all requested folders and their subfolders
		$AllFilesCount = 0				# Running total of the total number of files deleted
		$RemovedFolderCount = 1			# Some folders may not be removable so this is used to compare against the total number of folders
		$StartTime = (Get-Date)
		$DeleteParentFolder = $true		# If any files are not removable by Robocopy, this is disabled so the final removal of that folder is not attempted
        $TargetObject = $null			# Used to store any specific folders that give errors for display in log files
		#EndRegion Variable Declaration
		
		# The script is best run as an adminsitrator since the fewer access denieds the better
		# This section is used to check if "Run as Administrator" was used and if not to tell the user exit the script
		if(!([Security.Principal.WindowsPrincipal] [Security.Principal.WindowsIdentity]::GetCurrent()).IsInRole([Security.Principal.WindowsBuiltInRole]::Administrator))
		{ Write-Error "Script must be run with elevated permissions.  Run the script again as an administrator."; break }
		
		# If the user specifies whatif, we still have a number of new-item and remove-item lines that must be run for temp files
		# Therefore we will record the users preference and then disable whatif and only check it on operations we know are dangerous
		$RequestedWhatIfPreference = $WhatIfPreference

		#Region Robocopy Configuration
		<#
		/MIR	MIRror a directory tree (When used with an empty source folder, it effectively deletes the target folder)
		/l		List only - don’t copy, timestamp or delete any files (Log option, no changes are made.  Effectively -"whatif")
		/nc		No Class - don’t log file classes (Removes the "ADDED FILE" status column since we don't care about it)
		/njh	No Job Header (Only show one line per file and no other details as not to interfere with our split commands)
		/njs	No Job Summary (Only show one line per file and no other details as not to interfere with our split commands)
		/bytes	Print sizes as bytes (Ensures all sizes are the same so we can easily perform addition operations on them)
		/r:1	Number of Retries (We only want it try once and if not move onto the next one)
		/w:1	Wait time between retries (We don't want to wait if an error is encountered)
		#>
		
		# Specify the settings used by Robocopy
		$RobocopyArgs = @("*.*", "/MIR", "/nc", "/njh", "/njs", "/bytes", "/r:1", "/w:1")
		
		# If Whatif is enabled, add /l (Log only) to the robocopy paramaters and add text to the "WhatifPrefix variable so it appears in all logs
		if($RequestedWhatIfPreference) 
		{ 
			$RobocopyArgs += ("/l")
			$WhatIfPrefix = "[Whatif]"  
		}
		#EndRegion Robocopy Configuration
		
		# Disable Whatif as we are going to take control and only enable it where we know we are completing a dangerous operation
		$WhatIfPreference = $false

		#Region LogFile Validation 
		# If the user has specifed to enable writing to a log file
		if($LogFile)
		{
			try
			{
				# If the user writes to c:\temp\doesnotexist\log1.txt, make sure doesnotexist exists
				$LogParentPath = (Split-Path $LogFile)
				if(!(Test-Path($LogParentPath) -ErrorAction Stop)) {throw "Parent folder $LogParentPath does not exist."  }
				
				# Check if path is absolute (ie "c:\temp\log1.txt" and not simply "log1.txt")
				if(!([System.IO.Path]::IsPathRooted($LogFile))) {throw "$LogFile is not a complete or valid path to a file." }
				if((Get-Item $LogFile -ErrorAction SilentlyContinue) -is [System.IO.DirectoryInfo]) { throw "Invalid Logfile.  $LogFile is a directory.  Provide an absolute path and file name to the desired log file and try again."}
				if((Test-Path $LogFile -ErrorAction Stop)) 
				{
					$Prompt_Append = New-Object System.Management.Automation.Host.ChoiceDescription "&Append", "Appends to the existing log."
					$Prompt_Overwrite = New-Object System.Management.Automation.Host.ChoiceDescription "&Overwrite", "Overwrite the log."
					$options = [System.Management.Automation.Host.ChoiceDescription[]]($Prompt_Append, $Prompt_Overwrite)

					# If the user hasn't overridden this safety check with the -force paramater
					if(!($Force))
					{
						$LogChoice = $host.ui.PromptForChoice("", "`r`nLog File already exists.  Do you want to append or overwrite the file?", $options, 0) 
				
						switch ($LogChoice)
		    			{
					        0 { } # Append.  As we use add-content, append is the default action 
		        			1 { Remove-Item $LogFile -Force -ErrorAction Stop } # Overwrite (aka Delete and add to new)
		    			}
					}	
				}
				# If we are here, the log path is valid so test to make sure we have access to the file.  If not, it'll be caught in the try/catch
				else
				{ [io.file]::OpenWrite($LogFile).close(); Remove-Item $LogFile -Force -ErrorAction Stop }

			}
			Catch [System.Management.Automation.Host.PromptingException] { break }
			Catch { write-host $_ "Line" $($_.InvocationInfo.ScriptLineNumber) -ForegroundColor red; break }
		}
		#EndRegion LogFile Validation

		$RobocopyExe = "c:\Windows\System32\robocopy.exe"
		try { &$RobocopyExe | Out-Null } catch [System.Management.Automation.CommandNotFoundException] { Write-Error "Robocopy.exe cannot be found in c:\windows\system32" -ErrorAction Stop -ForeGroundColor Red }
		
		if(!($SkipRobocopyVersionCheck))
		{
			# Windows and Powershell report different versions for files in many cases so we have to do some added work to get the specific version
			# We basically want to make sure that the version of robocopy they have is the same one we tested with
			# If not, it'll likely still work but we want to make sure the user is aware that it is untested
			$VersionInfo = (Get-Item $RobocopyEXE).VersionInfo
			$RobocopyVersion = ("{0}.{1}.{2}.{3}" -f $VersionInfo.FileMajorPart, $VersionInfo.FileMinorPart, $VersionInfo.FileBuildPart, $VersionInfo.FilePrivatePart)
			
			switch ($RobocopyVersion)
			{
				'6.3.9600.17111' { } # Windows 8.1
				'6.3.9600.16384' { } # Windows 2012 R2 / 8.1 (earlier patches)
				'5.1.10.1027' { }    # Windows 7 SP1
				default { Write-Error "The version of robocopy on this system has not been tested with this script.  If you'd like to use it anyway, use the -skipRobocopyVerisonCheck paramater" -ErrorAction Stop }
			}
		}
		# We delete files by mirroing an empty folder with robocopy.  This requires a guaranteed empty folder.  This generates one for us
		Get-Random -Count 10 -InputObject (65..90) | % -begin {$EmptyFolder=$null} -process {$EmptyFolder += [char]$_} -End {$EmptyFolder = "$Env:TEMP\$EmptyFolder.emptyfolder" }
	}
	
	PROCESS 
	{ 
		# As we accept pipeline input, we can potentially get anything.  By getting the object type, we can ensure we only work on data we know how to use
		filter Get-TypeName { $_.GetType().Name}
		$InputObjectType = $InputFolder | Get-TypeName
		
		# If any of the provided folders are blank (typically extra lines in a CSV), discard them.  Return operates effectively as a continue in a PROCESS block
		if(!($InputFolder)) { return }

        # We use a switch so we can easily add more supported object types in the future if we later find them necessary
		switch($InputObjectType)
		{
			'String' 
			{
				# If an object comes in as a paramater, the .count will return the total number of entries
				# If an object comes from the pipleine, .count will always return 1 since in the PROCESS block we get each row one at a time
				# We need to know the total number of requested folders regardless if they came from the pipeline or a paramater
				# That's why we add them into the ArrayFolder, one at a time if from the pipeline and the entire object if from the paramater
				switch(($InputFolder).count)
				{
					'0' { Write-Error "Problem" ; break }
					'1' { $ArrayFolder += $InputFolder.trim() ; break } # provided Via Pipeline
					{'$_ -gt 1'} { $ArrayFolder = $InputFolder; break } # Provided via Paramater
				}
			}
			default { Write-Error "Input object '$InputObjectType' is not supported. Provide one or more folder paths as strings and try again." -ErrorAction Stop } 
		}
	}
	
	END
	{
		Try 
		{ 
			$FolderCount = $ArrayFolder.count
			# Create the empty folder we need to use to delete data with robocopy via mirroring.  This will be caught in the try/catch if there is an error
			New-Item -ItemType Directory -Path $EmptyFolder | Out-Null  
		
			# Write start time and the following message only to the log file rather than to the screen as well
			$LogMessage = "$WhatifPrefix`Removing the following files and folders:`r`n"
			if ($LogFile) { Add-Content -Path $LogFile -Value "[StartTime: $StartTime]" -ErrorAction Stop }
			if ($LogFile) { Add-Content -Path $LogFile -Value $LogMessage -ErrorAction Stop }
			
			ForEach($Folder in $ArrayFolder)
			{
				# Test if we have access to each folder by trying to write and delete a random file into it
				# If we don't have access, log the access issue and continue on to the next one
				if(!(Test-FolderAccess $Folder))
				{ 
					$LogMessage = "$Folder is not accessible"
					Write-Host $LogMessage -ForegroundColor Red
					if ($LogFile) { Add-Content -Path $LogFile -Value $LogMessage -ErrorAction Stop }
					continue
				}
				
				# Reset the intention to delete the parent folder on each execution
				$DeleteParentFolder = $True
				$FolderSize = 0
				$FileCount = 0
				# Checks if the folder exists -- the IsData filter solve for cases where $Folder is $null or blank
				filter IsData($var){if($var){$var}else{$false}}
				if(!(Test-Path ((IsData($Folder))) -ErrorAction Stop)) { $BadFolders += $Folder; continue }
				
				# If -force is not enabled, ask the user to confirm the deletion of the folder since this is where we are going to make changes
				if($Force -Or $PSCmdlet.ShouldContinue("$WhatifPrefix`Are you sure you want to remove '$Folder' ?", "", [ref]$ConfirmAll, [ref]$RejectAll)) 
				{
					# In order to get PowerShell to display the individual files deleted as well as to display a progress bar we use the pipeline
					# We pipe each line robocopy works on and parse the results.  This works by restricting what robocopy displays through its paramaters
					&$RoboCopyExe $EmptyFolder $Folder $RobocopyArgs  | Where-Object { $_} | select -Skip 1 | Foreach-Object `
					{ 
						#Store the current line returned by robocopy.  It should be in the format 1231 [tab] c:\temp\file1.txt
						# If there are any errors returned however, this format will not be met so we capture the line so we can send it directly as an error
						$InputLine = $_
						try { 
							# This double typecast will fail if the line is an error which allows us to record the error and move on
							$FileSize = [double](($_.trim()).split([char]9))[0]
							$FilePath = (($_.trim()).split([char]9))[1]
							$FileCount++
							# Robocopy returns directories with a size of -1.  We don't want to include those in our totals
							if($FileSize -ne $IsFolder) { $FolderSize = $FolderSize + $FileSize }
							$FileSizeInMB = [Math]::Round("{0:N0}" -f $FolderSize/1MB, 1)
							$AllFoldersSizeInMB = [Math]::Round("{0:N0}" -f $AllFoldersSize/1MB, 1)
							Write-Progress -Activity "[Started: $StartTime] ($RemovedFolderCount of $FolderCount) ($FileSizeInMB MB / $AllFoldersSizeinMB MB deleted)" -CurrentOperation "Current Folder: '$Folder'" -PercentComplete (($RemovedFolderCount*100)/$FolderCount)
							$LogMessage = "$WhatIfPrefix$FilePath"
							Write-Host $LogMessage
							if ($LogFile) { Add-Content -Path $LogFile -Value $LogMessage -ErrorAction Stop }
						}
						catch 
						{ 
							Write-Host $InputLine -ForegroundColor Red
							if ($LogFile) { Add-Content -Path $LogFile -Value $InputLine -ErrorAction Stop }
							# Something went wrong and we couldn't remove all of the data in the folder so don't bother trying to remove the parent folder
							$DeleteParentFolder = $False
						}
					}
			
					$RemovedFolderCount++
					$AllFoldersSize = $AllFoldersSize + $FolderSize
					$AllFilesCount = $AllFilesCount + $FileCount
					$FolderSize = [Math]::Round("{0:N0}" -f $FolderSize/1MB, 1)
					
					$LogMessage = "$WhatIfPrefix`Statistics for $Folder`: $FolderSize MB removed ($FileCount items)`r`n"
					Write-Host $LogMessage
					if ($LogFile) { Add-Content -Path $LogFile -Value $LogMessage -ErrorAction Stop }
					
					if($RequestedWhatIfPreference -eq $false -and $DeleteParentFolder) { Remove-Item $Folder -ErrorAction Stop }
				}
			}
			
			# We start $RemovedFolderCount at 1 so it looks right in the progress bar 
			# But this means that if there are two folders, it will run 3 times and so there will be 3 out 2 folders.  This subtraction corrects that
			$RemovedFolderCount--
			$AllFoldersSize = [Math]::Round("{0:N0}" -f $AllFoldersSize/1MB, 1)
			
			$LogMessage = "$WhatIfPrefix`Overall Statistics: $AllFilesCount files removed from $RemovedFolderCount of $FolderCount requested folders totaling $AllFoldersSize MB"
			Write-Host $LogMessage
			if ($LogFile) { Add-Content -Path $LogFile -Value $LogMessage -ErrorAction Stop }
			
			# Write any folders that could not be found both to the screen and to the log file if requested
			if($BadFolders)
			{
				$LogMessage = "$WhatIfPrefix`Folders Not Found: $BadFolders"
				Write-Warning  $LogMessage
				if ($LogFile) { Add-Content -Path $LogFile -Value $LogMessage -ErrorAction Stop }
			}
			
			# Now that we're done, we have to clean up and remove our temporary EmptyFolder from the Temp folder
			Remove-Item $EmptyFolder -ErrorAction Stop 
		
		} 
		# This PromptingException is for if the user presses cancel on any of the prompts.  We want the program to stop if they do that
		Catch [System.Management.Automation.Host.PromptingException] { }
		Catch 
        { 
            if($_.TargetObject) { [string]$TargetObject = $_.TargetObject }
			$LogMessage = "(TargetObject: $TargetObject) $_ Line $($_.InvocationInfo.ScriptLineNumber)"
			write-host $LogMessage -ForegroundColor red
            if ($LogFile) { Add-Content -Path $LogFile -Value $LogMessage -ErrorAction Stop }
        }
	}
}